# Ragel
# =====
#
# Utilities for the Ragel state machine compiler.
#
# This file is a part of BioJulia.
# License is MIT: https://github.com/BioJulia/BioCore.jl/blob/master/LICENSE.md

module Ragel

export tryread!

import BioCore.IO: FileFormat, AbstractReader, stream
import BufferedStreams: BufferedStreams, BufferedInputStream
#using Nullables

# A type keeping track of a ragel-based parser's state.
mutable struct State{T<:BufferedInputStream}
    stream::T      # input stream
    cs::Int        # current DFA state of Ragel
    linenum::Int   # line number: parser is responsible for updating this
    finished::Bool # true if finished (regardless of where in the stream we are)
end

function State(initstate::Int, input::BufferedInputStream)
    return State(input, initstate, 1, false)
end

#=
@inline function anchor!(stream, p)
    stream.anchor = 1 + p
end

@inline function upanchor!(stream)
    @assert stream.anchor != 0 "upanchor! called with no anchor set"
    anchor = stream.anchor
    stream.anchor = 0
    return anchor
end

# Parse an integer from an interval in the input buffer. This differs from
# `parse(Int, ...)` in that we don't have to copy or allocate anything, and it
# doesn't check that the characters are digits (we don't need to since this is
# already checked during parsing).
function parse_int64(buffer, firstpos, lastpos)
    if buffer[firstpos] == UInt8('-')
        sign = -1
        firstpos += 1
    else
        sign = +1
    end
    x = Int64(0)
    for i in firstpos:lastpos
        x = x * 10 + buffer[i] - UInt8('0')
    end
    return sign * x
end


# Actions
# -------

# Macros that help make common parsing tasks more succinct

macro anchor!()
    quote
        anchor!($(esc(:stream)), $(esc(:p)))
    end
end

macro copy_from_anchor!(dest)
    quote
        firstpos = upanchor!($(esc(:stream)))
        copy!($(esc(dest)), $(esc(:stream)).buffer, firstpos, $(esc(:p)))
    end
end

macro append_from_anchor!(dest)
    quote
        firstpos = upanchor!($(esc(:stream)))
        append!($(esc(dest)), $(esc(:stream)).buffer, firstpos, $(esc(:p)))
    end
end

macro int64_from_anchor!()
    quote
        firstpos = upanchor!($(esc(:stream)))
        parse_int64($(esc(:stream)).buffer, firstpos, $(esc(:p)))
    end
end

macro float64_from_anchor!()
    quote
        firstpos = upanchor!($(esc(:stream))) - 1
        get(ccall(
            :jl_try_substrtod,
            Nullable{Float64},
            (Ptr{UInt8}, Csize_t, Csize_t),
            $(esc(:stream)).buffer, firstpos, $(esc(:p)) - firstpos
        ))
    end
end

macro ascii_from_anchor!()
    quote
        firstpos = upanchor!($(esc(:stream)))
        n = $(esc(:p)) - firstpos + 1
        dst = Vector{UInt8}(n)
        copy!(dst, 1, $(esc(:stream)).buffer, firstpos, n)
        String(dst)
    end
end

macro load_from_anchor!(T)
    quote
        firstpos = upanchor!($(esc(:stream)))
        unsafe_load(convert(Ptr{$(T)},
            pointer($(esc(:stream)).buffer, firstpos)))
    end
end

# return the current character
macro char()
    quote
        convert(Char, $(esc(:stream)).buffer[$(esc(:p))+1])
    end
end

# advance `p`, set the current state to the target state `ts` and escape from
# scanning input
macro yield(ts)
    esc(quote
        p += 1
        cs = $ts
        @goto yield
    end)
end


# read/read! function generators
# ------------------------------

# Define a read!/read function wrapping ragel-generated parser.  These macros
# handle some the dirty work of maintaining state, refilling buffers, etc.
macro generate_read!_function(machine_name, input_type, output_type, ragel_body)
    accept_state = Symbol(machine_name, "_first_final")
    error_state = Symbol(machine_name, "_error")

    return esc(quote
        function Base.read!(input::$(input_type), output::$(output_type))
            return _read!(input, input.state, output)
        end

        function _read!(input::$(input_type), state, output::$(output_type))
            stream = state.stream

            if state.finished
                throw(EOFError())
            end

            # restore state variables used by Ragel (`p` is 0-based)
            cs = state.cs
            p = stream.position - 1
            pe = stream.available
            eof = -1
            data = stream.buffer

            if !isopen(stream)
                error("input stream is already closed")
            end

            if p ≥ pe
                # fill data buffer
                if BufferedStreams.fillbuffer!(state.stream) == 0
                    state.finished = true
                    throw(EOFError())
                end
                p = stream.position - 1
                pe = stream.available
            end

            # run the parser until all input is consumed or a match is found
            while true
                $(ragel_body)

                state.cs = cs
                stream.position = p + 1

                if cs == $(error_state)
                    error("error parsing ", $(machine_name),
                          " input on line ", state.linenum)
                elseif p == eof || state.finished  # exhausted all input data
                    if cs >= $(accept_state)
                        state.finished = true
                        throw(EOFError())
                    else
                        error("unexpected end of input while parsing ", $(machine_name))
                    end
                elseif p == pe  # exhausted filled input data
                    # refill data buffer
                    hits_eof = BufferedStreams.fillbuffer!(stream) == 0
                    p = stream.position - 1
                    pe = stream.available
                    if hits_eof
                        eof = pe
                    end
                else
                    # unreachable here
                    @assert false
                end
            end

            # `@yield` jumps here
            @label yield

            # save the current state and the reading position
            state.cs = cs
            stream.position = p + 1

            return output
        end
    end)
end

function Base.eof(input::AbstractReader)
    return eof(stream(input))
end
=#


end # module Ragel
